"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _pool;
const worker_threads_1 = require("worker_threads"); // eslint-disable-line
const events_1 = require("events");
const async_hooks_1 = require("async_hooks");
const os_1 = require("os");
const path_1 = require("path");
const util_1 = require("util");
const assert_1 = __importDefault(require("assert"));
const hdr_histogram_js_1 = require("hdr-histogram-js");
const perf_hooks_1 = require("perf_hooks");
const hdr_histogram_percentiles_obj_1 = __importDefault(require("hdr-histogram-percentiles-obj"));
const common_1 = require("./common");
const package_json_1 = require("../package.json");
// TODO(addaleax): Undo when https://github.com/DefinitelyTyped/DefinitelyTyped/pull/44034 is released.
const worker_threads_2 = __importDefault(require("worker_threads")); // eslint-disable-line
const { receiveMessageOnPort } = worker_threads_2.default;
const cpuCount = (() => {
    try {
        return os_1.cpus().length;
    }
    catch (_a) {
        /* istanbul ignore next */
        return 1;
    }
})();
function onabort(abortSignal, listener) {
    if ('addEventListener' in abortSignal) {
        abortSignal.addEventListener('abort', listener);
    }
    else {
        abortSignal.on('abort', listener);
    }
}
class AbortError extends Error {
    constructor() {
        super('The task has been aborted');
    }
}
const kDefaultOptions = {
    filename: null,
    minThreads: Math.max(cpuCount / 2, 1),
    maxThreads: cpuCount * 1.5,
    idleTimeout: 0,
    maxQueue: Infinity,
    concurrentTasksPerWorker: 1,
    useAtomics: true
};
let taskIdCounter = 0;
// Extend AsyncResource so that async relations between posting a task and
// receiving its result are visible to diagnostic tools.
class TaskInfo extends async_hooks_1.AsyncResource {
    constructor(task, transferList, filename, callback, abortSignal) {
        super('Piscina.Task', { requireManualDestroy: false });
        this.workerInfo = null;
        this.callback = callback;
        this.task = task;
        this.transferList = transferList;
        this.filename = filename;
        this.taskId = taskIdCounter++;
        this.abortSignal = abortSignal;
        this.created = perf_hooks_1.performance.now();
        this.started = 0;
    }
    releaseTask() {
        const ret = this.task;
        this.task = null;
        return ret;
    }
    done(err, result) {
        this.runInAsyncScope(this.callback, null, err, result);
        this.emitDestroy(); // `TaskInfo`s are used only once.
    }
}
class WorkerInfo {
    constructor(worker, port, onMessage) {
        this.idleTimeout = null;
        this.lastSeenResponseCount = 0;
        this.worker = worker;
        this.port = port;
        this.port.on('message', (message) => this._handleResponse(message));
        this.onMessage = onMessage;
        this.taskInfos = new Map();
        this.sharedBuffer = new Int32Array(new SharedArrayBuffer(common_1.kFieldCount * Int32Array.BYTES_PER_ELEMENT));
    }
    destroy() {
        this.worker.terminate();
        this.clearIdleTimeout();
        for (const taskInfo of this.taskInfos.values()) {
            taskInfo.done(new Error('Terminating worker thread'));
        }
        this.taskInfos.clear();
    }
    clearIdleTimeout() {
        if (this.idleTimeout !== null) {
            clearTimeout(this.idleTimeout);
            this.idleTimeout = null;
        }
    }
    ref() {
        this.worker.ref();
        this.port.ref();
        return this;
    }
    unref() {
        this.worker.unref();
        this.port.unref();
        return this;
    }
    _handleResponse(message) {
        this.onMessage(message);
        if (this.taskInfos.size === 0) {
            // No more tasks running on this Worker means it should not keep the
            // process running.
            this.unref();
        }
    }
    postTask(taskInfo) {
        assert_1.default(!this.taskInfos.has(taskInfo.taskId));
        const message = {
            task: taskInfo.releaseTask(),
            taskId: taskInfo.taskId,
            filename: taskInfo.filename
        };
        try {
            this.port.postMessage(message, taskInfo.transferList);
        }
        catch (err) {
            // This would mostly happen if e.g. message contains unserializable data
            // or transferList is invalid.
            taskInfo.done(err);
            return;
        }
        taskInfo.workerInfo = this;
        this.taskInfos.set(taskInfo.taskId, taskInfo);
        this.ref();
        this.clearIdleTimeout();
        // Inform the worker that there are new messages posted, and wake it up
        // if it is waiting for one.
        Atomics.add(this.sharedBuffer, common_1.kRequestCountField, 1);
        Atomics.notify(this.sharedBuffer, common_1.kRequestCountField, 1);
    }
    processPendingMessages() {
        // If we *know* that there are more messages than we have received using
        // 'message' events yet, then try to load and handle them synchronously,
        // without the need to wait for more expensive events on the event loop.
        // This would usually break async tracking, but in our case, we already have
        // the extra TaskInfo/AsyncResource layer that rectifies that situation.
        const actualResponseCount = Atomics.load(this.sharedBuffer, common_1.kResponseCountField);
        if (actualResponseCount !== this.lastSeenResponseCount) {
            this.lastSeenResponseCount = actualResponseCount;
            let entry;
            while ((entry = receiveMessageOnPort(this.port)) !== undefined) {
                this._handleResponse(entry.message);
            }
        }
    }
    isRunningAbortableTask() {
        // If there are abortable tasks, we are running one at most per Worker.
        if (this.taskInfos.size !== 1)
            return false;
        const [[, task]] = this.taskInfos;
        return task.abortSignal !== null;
    }
}
class ThreadPool {
    constructor(publicInterface, options) {
        this.completed = 0;
        this.publicInterface = publicInterface;
        this.workers = [];
        this.taskQueue = [];
        this.runTime = hdr_histogram_js_1.build({ lowestDiscernibleValue: 1 });
        this.waitTime = hdr_histogram_js_1.build({ lowestDiscernibleValue: 1 });
        this.options = { ...kDefaultOptions, ...options };
        // The >= and <= could be > and < but this way we get 100 % coverage 🙃
        if (options.maxThreads !== undefined &&
            this.options.minThreads >= options.maxThreads) {
            this.options.minThreads = options.maxThreads;
        }
        if (options.minThreads !== undefined &&
            this.options.maxThreads <= options.minThreads) {
            this.options.maxThreads = options.minThreads;
        }
        this._ensureMinimumWorkers();
    }
    _ensureMinimumWorkers() {
        while (this.workers.length < this.options.minThreads) {
            this._onWorkerFree(this._addNewWorker());
        }
    }
    _addNewWorker() {
        const pool = this;
        const worker = new worker_threads_1.Worker(path_1.resolve(__dirname, 'worker.js'));
        const { port1, port2 } = new worker_threads_1.MessageChannel();
        const workerInfo = new WorkerInfo(worker, port1, onMessage);
        const message = {
            filename: this.options.filename,
            port: port2,
            sharedBuffer: workerInfo.sharedBuffer,
            useAtomics: this.options.useAtomics
        };
        worker.postMessage(message, [port2]);
        function onMessage(message) {
            const { taskId, result } = message;
            // In case of success: Call the callback that was passed to `runTask`,
            // remove the `TaskInfo` associated with the Worker, which marks it as
            // free again.
            const taskInfo = workerInfo.taskInfos.get(taskId);
            workerInfo.taskInfos.delete(taskId);
            pool._onWorkerFree(workerInfo);
            /* istanbul ignore if */
            if (taskInfo === undefined) {
                const err = new Error(`Unexpected message from Worker: ${util_1.inspect(message)}`);
                pool.publicInterface.emit('error', err);
            }
            else {
                taskInfo.done(message.error, result);
            }
            if (pool.options.useAtomics) {
                for (const workerInfo of pool.workers) {
                    workerInfo.processPendingMessages();
                }
            }
        }
        worker.on('message', (message) => {
            worker.emit('error', new Error(`Unexpected message on Worker: ${util_1.inspect(message)}`));
        });
        worker.on('error', (err) => {
            // In case of an uncaught exception: Call the callback that was passed to
            // `postTask` with the error, or emit an 'error' event if there is none.
            const taskInfos = [...workerInfo.taskInfos.values()];
            workerInfo.taskInfos.clear();
            // Remove the worker from the list and potentially start a new Worker to
            // replace the current one.
            this._removeWorker(workerInfo);
            this._ensureMinimumWorkers();
            if (taskInfos.length > 0) {
                for (const taskInfo of taskInfos) {
                    taskInfo.done(err, null);
                }
            }
            else {
                this.publicInterface.emit('error', err);
            }
        });
        workerInfo.unref();
        this.workers.push(workerInfo);
        return workerInfo;
    }
    _removeWorker(workerInfo) {
        workerInfo.destroy();
        this.workers.splice(this.workers.indexOf(workerInfo), 1);
    }
    _onWorkerFree(workerInfo) {
        if (this.taskQueue.length > 0) {
            const taskInfo = this.taskQueue.shift();
            const now = perf_hooks_1.performance.now();
            this.waitTime.recordValue(now - taskInfo.created);
            taskInfo.started = now;
            workerInfo.postTask(taskInfo);
            return;
        }
        if (workerInfo.taskInfos.size === 0 &&
            this.workers.length > this.options.minThreads) {
            workerInfo.idleTimeout = setTimeout(() => {
                assert_1.default.strictEqual(workerInfo.taskInfos.size, 0);
                if (this.workers.length > this.options.minThreads) {
                    this._removeWorker(workerInfo);
                }
            }, this.options.idleTimeout).unref();
        }
    }
    runTask(task, transferList, filename, abortSignal) {
        if (filename === null) {
            filename = this.options.filename;
        }
        if (typeof filename !== 'string') {
            return Promise.reject(new Error('filename must be provided to runTask() or in options object'));
        }
        let resolve;
        let reject;
        // eslint-disable-next-line
        const ret = new Promise((res, rej) => { resolve = res; reject = rej; });
        const taskInfo = new TaskInfo(task, transferList, filename, (err, result) => {
            this.completed++;
            if (taskInfo.started) {
                this.runTime.recordValue(perf_hooks_1.performance.now() - taskInfo.started);
            }
            if (err !== null) {
                reject(err);
            }
            else {
                resolve(result);
            }
        }, abortSignal);
        if (abortSignal !== null) {
            onabort(abortSignal, () => {
                // Call reject() first to make sure we always reject with the AbortError
                // if the task is aborted, not with an Error from the possible
                // thread termination below.
                reject(new AbortError());
                if (taskInfo.workerInfo !== null) {
                    // Already running: We cancel the Worker this is running on.
                    this._removeWorker(taskInfo.workerInfo);
                    this._ensureMinimumWorkers();
                }
                else {
                    // Not yet running: Remove it from the queue.
                    const index = this.taskQueue.indexOf(taskInfo);
                    assert_1.default.notStrictEqual(index, -1);
                    this.taskQueue.splice(index, 1);
                }
            });
        }
        // If there is a task queue, there's no point in looking for an available
        // Worker thread. Add this task to the queue, if possible.
        if (this.taskQueue.length > 0) {
            if (this.taskQueue.length >= this.options.maxQueue) {
                return Promise.reject(new Error('Task queue is at limit'));
            }
            else {
                this.taskQueue.push(taskInfo);
            }
            return ret;
        }
        // Look for a Worker with a minimum number of tasks it is currently running.
        let workerInfo = null;
        let minimumCurrentTasks = Infinity;
        for (const candidate of this.workers) {
            if (candidate.taskInfos.size < minimumCurrentTasks &&
                !candidate.isRunningAbortableTask()) {
                workerInfo = candidate;
                minimumCurrentTasks = candidate.taskInfos.size;
                if (minimumCurrentTasks === 0)
                    break;
            }
        }
        // If all Workers are at maximum usage, do not use any of them.
        // If we want the ability to abort this task, use only workers that have
        // no running tasks.
        if (minimumCurrentTasks >= this.options.concurrentTasksPerWorker ||
            (minimumCurrentTasks > 0 && abortSignal)) {
            workerInfo = null;
        }
        // If no Worker was found, or that Worker was handling another task in some
        // way, and we still have the ability to spawn new threads, do so.
        if ((workerInfo === null || minimumCurrentTasks > 0) &&
            this.workers.length < this.options.maxThreads) {
            // This doesn't account for Worker startup time yet. If another Worker
            // becomes available beefore the new one is up, we should use that.
            workerInfo = this._addNewWorker();
        }
        // If no Worker is found, try to put the task into the queue.
        if (workerInfo === null) {
            if (this.options.maxQueue <= 0) {
                return Promise.reject(new Error('No task queue available and all Workers are busy'));
            }
            else {
                this.taskQueue.push(taskInfo);
            }
            return ret;
        }
        const now = perf_hooks_1.performance.now();
        this.waitTime.recordValue(now - taskInfo.created);
        taskInfo.started = now;
        workerInfo.postTask(taskInfo);
        return ret;
    }
    async destroy() {
        while (this.taskQueue.length > 0) {
            const taskInfo = this.taskQueue.shift();
            taskInfo.done(new Error('Terminating worker thread'));
        }
        const exitEvents = [];
        while (this.workers.length > 0) {
            exitEvents.push(events_1.once(this.workers[0].worker, 'exit'));
            this._removeWorker(this.workers[0]);
        }
        await Promise.all(exitEvents);
    }
}
class Piscina extends events_1.EventEmitter {
    constructor(options = {}) {
        super(options);
        _pool.set(this, void 0);
        if (typeof options.filename !== 'string' && options.filename != null) {
            throw new TypeError('options.filename must be a string or null');
        }
        if (options.minThreads !== undefined &&
            (typeof options.minThreads !== 'number' || options.minThreads < 0)) {
            throw new TypeError('options.minThreads must be a non-negative integer');
        }
        if (options.maxThreads !== undefined &&
            (typeof options.maxThreads !== 'number' || options.maxThreads < 1)) {
            throw new TypeError('options.maxThreads must be a positive integer');
        }
        if (options.minThreads !== undefined && options.maxThreads !== undefined &&
            options.minThreads > options.maxThreads) {
            throw new RangeError('options.minThreads and options.maxThreads must not conflict');
        }
        if (options.idleTimeout !== undefined &&
            (typeof options.idleTimeout !== 'number' || options.idleTimeout < 0)) {
            throw new TypeError('options.idleTimeout must be a non-negative integer');
        }
        if (options.maxQueue !== undefined &&
            (typeof options.maxQueue !== 'number' || options.maxQueue < 0)) {
            throw new TypeError('options.maxQueue must be a non-negative integer');
        }
        if (options.concurrentTasksPerWorker !== undefined &&
            (typeof options.concurrentTasksPerWorker !== 'number' ||
                options.concurrentTasksPerWorker < 1)) {
            throw new TypeError('options.concurrentTasksPerWorker must be a positive integer');
        }
        if (options.useAtomics !== undefined &&
            typeof options.useAtomics !== 'boolean') {
            throw new TypeError('options.useAtomics must be a boolean value');
        }
        __classPrivateFieldSet(this, _pool, new ThreadPool(this, options));
    }
    runTask(task, transferList, filename, abortSignal) {
        // If transferList is a string or AbortSignal, shift it.
        if ((typeof transferList === 'object' && !Array.isArray(transferList)) ||
            typeof transferList === 'string') {
            abortSignal = filename;
            filename = transferList;
            transferList = undefined;
        }
        // If filename is an AbortSignal, shift it.
        if (typeof filename === 'object' && !Array.isArray(filename)) {
            abortSignal = filename;
            filename = undefined;
        }
        if (transferList !== undefined && !Array.isArray(transferList)) {
            return Promise.reject(new TypeError('transferList argument must be an Array'));
        }
        if (filename !== undefined && typeof filename !== 'string') {
            return Promise.reject(new TypeError('filename argument must be a string'));
        }
        if (abortSignal !== undefined && typeof abortSignal !== 'object') {
            return Promise.reject(new TypeError('abortSignal argument must be an object'));
        }
        return __classPrivateFieldGet(this, _pool).runTask(task, transferList, filename || null, abortSignal || null);
    }
    destroy() {
        return __classPrivateFieldGet(this, _pool).destroy();
    }
    get options() {
        return __classPrivateFieldGet(this, _pool).options;
    }
    get threads() {
        return __classPrivateFieldGet(this, _pool).workers.map((workerInfo) => workerInfo.worker);
    }
    get queueSize() {
        return __classPrivateFieldGet(this, _pool).taskQueue.length;
    }
    get completed() {
        return __classPrivateFieldGet(this, _pool).completed;
    }
    get waitTime() {
        const result = hdr_histogram_percentiles_obj_1.default.histAsObj(__classPrivateFieldGet(this, _pool).waitTime);
        return hdr_histogram_percentiles_obj_1.default.addPercentiles(__classPrivateFieldGet(this, _pool).waitTime, result);
    }
    get runTime() {
        const result = hdr_histogram_percentiles_obj_1.default.histAsObj(__classPrivateFieldGet(this, _pool).runTime);
        return hdr_histogram_percentiles_obj_1.default.addPercentiles(__classPrivateFieldGet(this, _pool).runTime, result);
    }
    static get isWorkerThread() {
        return common_1.commonState.isWorkerThread;
    }
    static get version() {
        return package_json_1.version;
    }
    static get Piscina() {
        return Piscina;
    }
}
_pool = new WeakMap();
module.exports = Piscina;
//# sourceMappingURL=index.js.map